import copy
import importlib
from typing import *

__all__ = [
    "get_object_from_module",
    "instantiate_class",
    "build_instance_from_dict",
    "get_progress_bar",
]

VALUE_TYPES = (int, float, str, bool)
LIST_TYPES = (list, set, tuple)
DICT_TYPES = (dict,)


def get_object_from_module(module_path, object_name):
    m = importlib.import_module(module_path)
    c = getattr(m, object_name)
    return c


def instantiate_class(info: Dict):
    if "kwargs" in info:
        kwargs = info["kwargs"]
    else:
        kwargs = {}

    assert "module" in info, f"Key `module` not in {info}"
    assert "class" in info, f"Key `class` not in {info}"

    return get_object_from_module(info["module"], info["class"])(**kwargs)


def build_instance_from_dict(
    config: Dict, source_key: str, default_key: str = "default_params"
):
    """
    Params:
        config: Dict, config of the instance with keys ('instance_name_1', 'instance_name_2', default_key)
        source_key: str, name of the instance
        default_key: str, name of the key of default params in config
    """

    cfg = copy.deepcopy(config)

    default = cfg[default_key] if default_key in cfg else {}
    for k, v in default.items():
        assert v is not None, f"Default params should not be empty but {k} is empty."
    src = cfg[source_key]

    def replace(v, key=None):
        # replace None by default value
        # replace config_dict by python class

        if v is None:
            if key is None or key not in default:
                raise ValueError(
                    f"None type found with key {key} without replaceable default."
                )
            return replace(default[key])

        if isinstance(v, VALUE_TYPES):
            return v

        if isinstance(v, DICT_TYPES):
            for k, w in v.items():
                v[k] = replace(w, key=k)
            if "module" in v and "class" in v:
                return instantiate_class(v)
            return v

        if isinstance(v, LIST_TYPES):
            out = []
            for x in v:
                out.append(replace(x))
            return out

    if "kwargs" not in src:
        return instantiate_class(src)

    return replace(src)


def get_progress_bar(itr: int, tot: int, width: int = 25):

    done = int(itr / tot * width)
    todo = width - done
    progress_bar = "[" + "=" * (done) + " " * (todo) + "]"

    out = []
    out.append(f">> {itr+1}|{tot}")
    out.append(progress_bar)
    out.append(f"{int((itr+1)/tot * 100)}%")
    out.append(" " * 10)

    out = "  ".join(out)

    return out



class BreakIteration(Exception):
    def __init__(self, shutdown_engine=False):
        self.shutdown_engine = shutdown_engine


class ContinueIteration(Exception):
    pass


class BadBatchError(ContinueIteration):
    """Use case: current batch of data is corrupted, skip current batch and fetch a new batch and then continue."""

    pass


class StopEngine(BreakIteration):
    """Use case: when raised at batch iteration level, if `shutdown_engine=True` then reraise BreakIteration to terminate engine."""

    def __init__(self):
        super().__init__(True)


class GradientExplosionError(StopEngine):
    pass


class FoundNanError(StopEngine):
    pass


class EarlyStopping(StopEngine):
    pass



import collections
import warnings


from .component_types import *
from .. import REPR_INDENT
from ..exception import BreakIteration, ContinueIteration


__all__ = ["EngineBase"]


class EngineBase:

    _registry = tuple([f"{c}_registry" for c in map_str_to_types])

    def __init__(self):
        for name in self._registry:
            super().__setattr__(name, {})
        self.reset_engine()

    def reset_engine(self):
        self.epoch = 0
        self.iteration = 0
        self.epoch_length = None
        self.stubs_in_queue = collections.deque()
        self.stubs_done = []
        self.current_stub = None

    def __setattr__(self, name, value):
        """
        set attribute operator.

        - if name exists
          - if current attr is engine component: raise current attr is a component please del it first
          - if current attr is a registry: raise registry is immutable
          - else: treat it as a new attr
        - if name is new
          - if value is a component: register in registry
          - else: pass
          - super() set attr
        """

        if hasattr(self, name):

            if isinstance(getattr(self, name), COMPONENTS):
                raise AttributeError(
                    f"Attribute `{name}` is an engine component. You need to delete it first before assigning a new value."
                )

            if name == "_registry":
                raise AttributeError(f"`_registry` is a protected attribute.")

            if name in self._registry:
                raise AttributeError(f"`{name}` is immutable.")

        if isinstance(value, COMPONENTS):

            for TYPE in COMPONENTS:
                if isinstance(value, TYPE):
                    typestr = map_types_to_str[TYPE]
                    break
            registry = getattr(self, f"{typestr}_registry")
            assert (
                name not in registry
            ), f"The `{name}` is already in `{typestr}_registry`. Components should be registered and deregistered through setting and deleting attribues."
            registry[name] = value

        super().__setattr__(name, value)

    def __delattr__(self, name):
        """delete attribute operator if attr is a component, also deregister the
        component in the corresponding registry."""

        value = getattr(self, name)

        if isinstance(value, COMPONENTS):

            for TYPE in COMPONENTS:
                if isinstance(value, TYPE):
                    typestr = map_types_to_str[TYPE]
                    break
            registry = getattr(self, f"{typestr}_registry")
            del registry[name]

        super().__delattr__(name)

    def state_dict(self):
        out = {}

        out["engine"] = {"epoch": self.epoch, "iteration": self.iteration}

        for rname in self._registry:
            registry = getattr(self, rname)
            cname = rname.split("_")[0]  # e.g. `model`
            if (
                cname == "dataloader"
            ):  # current dataloader class doesn't have state dict
                continue
            out[cname] = {k: v.state_dict() for k, v in registry.items()}

        return out

    def load_state_dict(self, state_dict):

        if "engine" not in state_dict:
            self.reset_engine()

        for cname, cstate in state_dict.items():

            if cname == "engine":
                epoch = int(state_dict["engine"]["epoch"])
                iteration = int(state_dict["engine"]["iteration"])
                if epoch < 0 or iteration < 0:
                    self.reset_engine()
                    warnings.warn("Invalid values encountered in engine state_dict.")
                else:
                    self.epoch = epoch
                    self.iteration = iteration
                continue

            # otherwise
            registry = getattr(self, f"{cname}_registry")
            for k, v in cstate.items():
                registry[k].load_state_dict(v)

    def train(self):
        for k, v in self.model_registry.items():
            v.train()

    def eval(self):
        for k, v in self.model_registry.items():
            v.eval()

    @property
    def training(self):
        out = {}
        for k, v in self.model_registry.items():
            out[k] = v.training
        return out

    def per_batch(self, batch, **kwargs):
        raise NotImplementedError("Method `per_batch` is not implemented")

    def when_epoch_starts(self, **kwargs):
        pass

    def when_epoch_finishes(self, **kwargs):
        pass

    def between_iterations(self, **kwargs):
        pass

    def per_epoch(self, **kwargs):
        """Train model by one epoch."""

        if self.current_stub.dataloader in self.dataloader_registry:
            dataloader = getattr(self, self.current_stub.dataloader)
        else:
            raise AttributeError(
                f"`{self.current_stub.dataloader}` is not in dataloader registry."
            )

        self.epoch_length = len(dataloader)
        self.when_epoch_starts(**kwargs)
        try:
            # if current stub was paused, then pick up from there
            it = self.current_stub.iteration
            it = int(it)
            self.iteration = it if it > 0 else 0
        except AttributeError as e:
            # if stub does not have iteration
            warnings.warn(f"{e} Iteration reset to zero.")
            self.iteration = 0

        for batch in dataloader:
            try:
                self.per_batch(batch, **kwargs)
                self.iteration += 1
                self.between_iterations(**kwargs)
                if self.iteration == self.epoch_length:
                    # terminate such that total iterations equal epoch length
                    # NOTE: so far assume sampling with replacement which is a good approximiation if batch is large
                    # TODO: add sampling without replacement
                    raise BreakIteration(False)
            except ContinueIteration:
                continue
            except BreakIteration as e:
                if e.shutdown_engine:
                    raise BreakIteration
                break
            except Exception as e:
                raise e

        self.epoch += 1
        self.when_epoch_finishes(**kwargs)

    def queue(self, stubs):
        self.stubs_in_queue.extend(stubs)

    def execute(self, **kwargs):
        while len(self.stubs_in_queue) > 0:
            self.current_stub = self.stubs_in_queue.popleft()
            if self.current_stub.action == "train":
                try:
                    self.per_epoch(**kwargs)
                except BreakIteration:
                    break
                except ContinueIteration:
                    continue
            else:
                try:
                    action = getattr(self, self.current_stub.action)
                    action(**kwargs)
                except AttributeError as e:
                    warnings.warn(f"{e} Job aborted.")
                    continue
            self.stubs_done.append(self.current_stub)

    def __call__(self, stubs=None, **kwargs):
        if stubs:
            self.queue(stubs)
        elif len(self.stubs_in_queue) == 0:
            raise ValueError("No job in queue. Stubs should not be empty")

        try:
            self.execute(**kwargs)
        except KeyboardInterrupt:
            self.current_stub.iteration = self.iteration
            self.stubs_in_queue.appendleft(self.current_stub)

    def __repr__(self):
        out = []
        out.append(self.__class__.__name__)
        out.append(" " * REPR_INDENT + f"epoch: {self.epoch}")
        out.append(" " * REPR_INDENT + f"iteration: {self.iteration}")

        for registry in self._registry:
            out.append(" " * REPR_INDENT + f"{registry}: ")
            for k in getattr(self, registry):
                out.append(" " * 2 * REPR_INDENT + f"{k}")

        return "\n".join(out)



import abc
import collections
from functools import wraps

import numpy as np
import torch
import torch.nn.functional as F

from . import REPR_INDENT

__all__ = ["to_buffer", "BufferBase", "ScalarSmoother", "VectorSmoother"]


def to_buffer(name="buffer_registry"):
    # name should be an attribute of the owner class

    def decorator(func):
        # func: class method that yields tuple of (key: str, val)
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            buffer_dict = getattr(self, name)
            for key, val in func(self, *args, **kwargs):
                if key in buffer_dict:
                    buffer_dict[key](val)  # pushing update by `__call__`

        return wrapper

    return decorator


class BufferBase(abc.ABC):
    """Buffer base class."""

    def __init__(self, *args, **kwargs):
        assert len(args) == 0, "There should not be any args only kwargs allowed."
        for k, v in kwargs.items():
            setattr(self, k, v)
        self.reset()

    @abc.abstractmethod
    def update(self, x):
        pass

    @abc.abstractmethod
    def reset(self):
        pass

    @abc.abstractmethod
    def state_dict(self):
        pass

    @abc.abstractmethod
    def load_state_dict(self, state_dict):
        pass

    def __call__(self, x):
        self.update(x)

    def __repr__(self):
        out = []
        out.append(self.__class__.__name__)
        for k, v in self.__dict__.items():
            if k.startswith("_"):
                continue
            out.append(" " * REPR_INDENT + f"{k}: {v}")
        return "\n".join(out)


class ScalarSmoother(BufferBase):
    """Rolling smoothing buffer for scalars."""

    def __init__(self, window_size: int, **kwargs):

        window_size = int(window_size)
        assert window_size > 0, f"window_size should be > 0 but get {window_size}"

        super().__init__(window_size=window_size, **kwargs)

    def reset(self):
        self._count = 0
        self._queue = collections.deque([], maxlen=self.window_size)

    def update(self, x: float):
        self._queue.append(x)
        self._count += 1

    def state_dict(self):
        return {"queue": self._queue, "count": self._count}

    def load_state_dict(self, state_dict):
        self._count = state_dict["count"]
        self._queue = state_dict["queue"]

    @property
    def mean(self):
        return np.mean(self._queue) if len(self._queue) > 0 else 0.0

    @property
    def median(self):
        return np.median(self._queue) if len(self._queue) > 0 else 0.0

    @property
    def std(self):
        return np.std(self._queue) if len(self._queue) > 0 else 0.0

    @property
    def max(self):
        return np.max(self._queue) if len(self._queue) > 0 else 0.0

    @property
    def min(self):
        return np.min(self._queue) if len(self._queue) > 0 else 0.0


class VectorSmoother(BufferBase):
    """
    Exponential moving average of n-dim vector:

    vector = alpha * new_vector + (1 - alpha) * vector
    """

    def __init__(
        self,
        alpha: float,
        n_dim: int,
        init_value: float,
        eps: float = 1e-8,
        normalize: bool = True,
        p: float = 1.0,
        device: str = "cpu",
        dtype: torch.dtype = torch.float,
        **kwargs,
    ):
        alpha = float(alpha)
        n_dim = max(1, int(n_dim))
        assert (
            alpha >= 0 and alpha <= 1
        ), f"Parameter alpha = {alpha} should be in [0, 1]"

        if dtype in (torch.int, torch.long):
            assert (
                not normalize
            ), f"If dtype is `int` or `long`, normalize must be `False`. "
        super().__init__(
            alpha=alpha,
            n_dim=n_dim,
            init_value=init_value,
            eps=eps,
            normalize=normalize,
            p=p,
            device=torch.device(device),
            dtype=dtype,
            **kwargs,
        )

    def reset(self):
        self._count = 0
        self._state = torch.zeros(self.n_dim) + self.init_value
        self._state = self._state.to(device=self.device, dtype=self.dtype)
        if self.normalize:
            self._state = self.lp_normalized(self.p)

    def update(self, x: torch.Tensor):
        self._state = self.alpha * x + (1 - self.alpha) * self._state
        if self.normalize:
            self._state = self.lp_normalized(self.p)

        self._count += 1

    def state_dict(self):
        return {"vector": self._state, "count": self._count}

    def load_state_dict(self, state_dict):
        self._count = state_dict["count"]
        self._state = state_dict["vector"]

    @property
    def vector(self):
        return self._state

    @property
    def l1_normalized(self):
        return self.lp_normalized(1.0)

    @property
    def l2_normalized(self):
        return self.lp_normalized(2.0)

    @property
    def l1_norm(self):
        return self.lp_norm(1.0)

    @property
    def l2_norm(self):
        return self.lp_norm(2.0)

    def lp_norm(self, p: float):
        return self._state.norm(dim=0, p=p)

    def lp_normalized(self, p: float):
        return F.normalize(self._state, dim=0, p=p, eps=self.eps)
